/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "ranger.hh"
#include "psd_estimate.hh"
#include "DVecType.hh"
#include "window_api.hh"
#include "MonServer.hh"
#include "Trend.hh"
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <cmath>

using namespace std;
using containers::fSeries;
using containers::PSD;

//======================================  Calculate the red-shift.
static const double cinkms = 3e5; // c in km/s
double ranger::H0 = 73.8;         // hubble constant in km/s/Mpc

//  Calculate the red-shift (z) from a range in Mpc.
inline double
z_from_r(double r) {
   double beta = r * ranger::H0 / cinkms;
   if (beta < 0 || beta >= 1.0) {
      ostringstream msg;
      msg << "z_from_r: invalid r (" << r << ").";
      throw std::runtime_error(msg.str());
   }
   double gamma = sqrt((beta+1)/(1-beta));
   return gamma - 1;
}

static fSeries
red_shift_series(double z, const fSeries& psd) {
   if (z < 0.001) return psd;

   //------------------------------------  Create a bogus psd with wider samples
   double dF0 = psd.getFStep();
   double dFz = dF0 * (z + 1.0);
   fSeries zseries(0.0, dFz, psd.getStartTime(), psd.getDt(), psd.refDVect());
   return zseries.interpolate(0, psd.getHighFreq(), dF0, fSeries::kLog);
}

static double
integrate(const DVector& f, double dx) {
   size_t N = f.size();
   if (N < 2) {
      return 0.0;
   }
   double integral = f.getDouble(0) + f.getDouble(N-1);
   if (N == 2) {
      //---------------------------------  N=2 trapezoid rule
      integral /= 2;
   } if (N%2 == 0) {
      //---------------------------------  trapezoid rule
      integral = 0.5 *integral + f.VSum(1, N-2);
   } else {
      //---------------------------------  Simpsons rule
      integral += 4 * f.getDouble(N-2);
      for (auto i=1; i<N - 2; i+=2) {
	 integral += 4 * f.getDouble(i) + 2 * f.getDouble(i+1);
      }
      integral /= 3.0;
   }
   integral *= dx;
   return integral;
}

//=======================================  Ranger constructor
ranger::ranger(void)
   : mSigThreshold(8.0), mFmin(10.0), mFmax(1400.0),
      mRangeHisto("", 100, 100, 150), mStart(0), mCurrent(0)
{
}

//=======================================  Ranger constructor
ranger::ranger(const std::string& name)
   : mToolName(name), mSigThreshold(8.0), mFmin(10.0), mFmax(1400.0),
     mRangeHisto("", 100, 100, 150), mStart(0), mCurrent(0)
{
}

//=======================================  Ranger destructor
ranger::~ranger(void) {
   if (mPsdTool) mPsdTool->status(cout);
}

//=======================================  Book the trends.
void
ranger::bookTrends(Trend& tr) {
   if (!mRangeTrend.empty()) {
      tr.addChannel(mRangeTrend);
      tr.find(mRangeTrend).setUnits("Mpc");
   }
   if (!mRedShiftTrend.empty()) {
      tr.addChannel(mRedShiftTrend);
      tr.find(mRedShiftTrend).setUnits("NONE");
   }
}

//=======================================  Calculate range
void
ranger::calculate(const TSeries& data) {
   Time t0     = data.getStartTime();
   Interval dt = data.getInterval();
   if (!mStart) {
      mStart = t0;
      mCurrent = mStart;
      padHistory(t0, dt);
   } else if (mCurrent != t0) {
      mCurrent = t0;
      // any resets?
   }

   //----------------------------------- calculate the noise spectrum
   mPsdTool->reset_accumulators();
   mPsdTool->add(data);
   mRawPSD = mPsdTool->get_psd();
   calibrate();
   mNoiseASD = mCalibPSD;

   //-----------------------------------  Get the z=0 range
   double rangez = range(mWeightZ0, mIntegrand);
   double z = z_from_r(rangez);

   //-----------------------------------  Iterate over a fe
   if (mRedShift) {
      const double dzMax(0.05);
      for (int nIter=0; nIter < 10; nIter++) {
	 PSD psdi, snrspec;
	 static_cast<fSeries&>(psdi) = red_shift_series(z, mWeightZ0);
	 double rangei = range(psdi, snrspec);
	 double zi     = z_from_r(rangei);
	 if (abs(zi-z) < dzMax) break;
	 z      = zi;
	 rangez = rangei;
	 mIntegrand = snrspec;
      }
   }

   //----------------------------------- publish monitor data
   mRangeHistory.fixedAppend(t0, dt, &rangez, 1);
   mRedShiftHistory.fixedAppend(t0, dt, &z, 1);
   mRangeHisto.Fill(rangez, 1.0);

   mRange = rangez;
   mZ     = z;
   cout << "range for " << mToolName << " at t=" << data.getStartTime().getS()
	<< " is " << mRange << ", red-shift is " << z << endl;

   mLast    = mCurrent;
   mCurrent = data.getEndTime();
}

//======================================  Calibrate the PSD
void
ranger::calibrate(void) {
   mCalibPSD = mRawPSD;
}

//======================================  inline display
inline const char*
tf(bool torf) {
   return tf ? "true" : "false";
}
inline string
ps(const string& str) {
   return string("\"") + str + "\"";
}

//======================================  Display the ranger parameters
void
ranger::display(std::ostream& out) const {
   out << "Range tool: " << mToolName << endl;
   out << "  Range trend channel name:       " << ps(mRangeTrend) << endl;
   out << "  Range monitor channel name:     " << ps(mRangeMonitor) << endl;
   out << "  Red-shift trend channel name:   " << ps(mRedShiftTrend) << endl;
   out << "  Red-shift monitor channel name: " << ps(mRedShiftMonitor) << endl;
   out << "  PSD monitor channel name:       " << ps(mPsdMonitor) << endl;
   out << "  Integrand channel name:         " << ps(mIntegrandMonitor) << endl;
   out << "  Monitor History length:         " << mHistoryLength << endl;
   out << "  Enable red-shift calculation:   " << tf(mRedShift) << endl;

   //-----------------------------------  Psd type
   out << "  PSD estimator:                  " << psd_tool_type() << endl;
   out << "  Window type (parameter):        " << psd_window_type() << endl;
   out << "  Window length:                  " << mWindowLength << endl;
   out << "  Signal name/file:               " << mGWType << endl;
   out << "  Significance threshold:         " << mSigThreshold << endl;
   out << "  Minimum integration frequency:  " << mFmin << endl;
   out << "  Maximum integration frequency:  " << mFmax << endl;
}

//======================================  Enable red-shift calculations
void
ranger::enableRedShift(bool rs) {
   mRedShift = rs;
}

//=======================================  Fill the trend channels.
void
ranger::fillTrends(Trend& tr) {
   if (mRangeTrend.size())    tr.trendData(mRangeTrend,    mCurrent, mRange);
   if (mRedShiftTrend.size()) tr.trendData(mRedShiftTrend, mCurrent, mRedShift);
}

//======================================  Generate a gw template psd
const containers::PSD
ranger::get_template(const std::string& file) const {
   //-----------------------------------  Constants
   const double massSun = 1.989e30;  // Solar mass (in kg)
   const double G       = 6.673e-11; // Newton's grav constant (N m^2 / kg^2)
   const double c       = 2.998e8;   // speed of light (m/s)
   const double Mpc     = 3.086e22;  // Mpc in m
   const double rho0    = 8;         // What is this? significance threshold?
   const double Gc2MpsSm = 4.786e-20; // G/c^2 in Mpc / Msun
   //----------------------------------  integral from Finn(1993) Eq. (5.3)
   const double geomfact = pow(3*1.84, 2.0/3.0);

   //-----------------------------------  Get frequency spacing
   double dF   = 1.0 / mWindowLength;
   double fMax = mFmax;
   size_t nBin = fMax * mWindowLength + 1;
   double fny  = 8192;
   DVectD dvd(nBin, 0);
   if (file == "bns-1.4x1.4") {
      double m1 = 1.4, m2=1.4;
      double mTot = (m1 + m2) * massSun * G / (c * c * c);
      double eta = m1 * m2 / ((m1+m2) * (m1+m2));

      //  m_Chirp/c^2 is in units of kg
      double m_Chirp = pow(pow(m1*m2, 3)/(m1+m2), 0.2) * massSun * G / (c * c);
      //  r0^2 in units of Mpc^2,
      double r0sq = 5.0 * geomfact * pow(pow(m_Chirp, 5.) * c, 1./3.)
                  / (96.0 * pow(M_PI, 4./3.) * Mpc * Mpc);
      double f_isco = 1.0/(sqrt(216) * M_PI * mTot);
      cout << "f_isco: " << f_isco << endl;
      dvd[0] = 0;
      for (auto i=1; i<nBin; i++) {
	 double f = double(i) * dF;
	 if (f >= mFmin && f < f_isco) dvd[i] = r0sq / pow(f, 7./3.);
	 else                          dvd[i] = 0;
      }
   } else {
      throw runtime_error(string("Unable to create template for: ") + file);
   }
   PSD weight;
   static_cast<fSeries&>(weight)=fSeries(0, dF, Time(0), Interval(1.0/dF), dvd);
   return weight;
}

//======================================  Define monitor channel names
void
ranger::monitorDefaultPrefix(const std::string& prefix) {
   mRangeMonitor     = prefix + "_EFF_RANGE";
   mRedShiftMonitor  = prefix + "_RED_SHIFT";
   mPsdMonitor       = prefix + "_NOISE_ASD";
   mIntegrandMonitor = prefix + "_INTEGRAND";
   mRangeHistoName   = prefix + "_RANGE_HISTO";
}

//======================================  Set up monitor channels
void
ranger::monitorSetup(MonServer& mon) {
   mon.serveData(mRangeMonitor,     &mRangeHistory);
   mon.serveData(mRedShiftMonitor,  &mRedShiftHistory);
   mon.serveData(mPsdMonitor,       &mNoiseASD);
   mon.serveData(mIntegrandMonitor, &mIntegrand);
   mon.serveData(mRangeHistoName,   &mRangeHisto);
}

//======================================  Extend the history series to tEnd
void
ranger::padHistory(const Time& tEnd, Interval dT) {
   if (!dT) {
      dT = mRangeHistory.getTStep();
      if (!dT) throw logic_error("ranger: Invalid history increment");
   }
   mRangeHistory.padUntil(tEnd, dT);
   mRedShiftHistory.padUntil(tEnd, dT);
}

//======================================  psd tool type
string
ranger::psd_tool_type(void) const {
   return mPsdTool->get_type();
}

//======================================  psd window type
string
ranger::psd_window_type(void) const {
   const Pipe* wpipe = mPsdTool->ref_window();
   ostringstream window;
   window << window_type(wpipe);
   if (window.str() == "tukey") {
      window << " (" << dynamic_cast<const window_api*>(wpipe)->parameter()
	     << ")";
   }
   return window.str();
}

//======================================  Calculate the range.  
double
ranger::range(const containers::PSD& psd, containers::PSD& rMpc) const {
   // psd.dump_header(cout);
   // mCalibPSD.dump_header(cout);
   double dF = psd.getFStep();
   double ftop = mFmax + dF;
   rMpc = psd;
   rMpc /= mCalibPSD.extract(0, mFmax);
   DVectD psd_band(rMpc.extract(mFmin, mFmax).refDVect());
   double rBand = integrate(psd_band, dF);
   if (rBand <= 0) return 0.0;
   return sqrt(rBand)/mSigThreshold;
}

//======================================  Set the minimum integration frequency
double
ranger::recent(Interval dt) const {
   if (mRangeHistory.empty()) return 0.0;
   Time start = mRangeHistory.getEndTime() - dt;
   TSeries recent_series = mRangeHistory.extract(start, dt);
   size_t nAvg = recent_series.getNGreater(0.0);
   if (!nAvg) return 0.0;
   size_t N = recent_series.getNSample();
   if (nAvg == N) return recent_series.getAverage();
   double sum = 0;
   int nsum = 0;
   for (size_t i=0; i<N; i++) {
      double x = recent_series.getDouble(i);
      if (x > 0) {
	 sum += x;
	 nsum++;
      }
   }
   if (!nsum) return 0.0;
   return sum/double(nsum);
}

//======================================  Set the minimum integration frequency
void
ranger::setFmin(double fmin) {
   mFmin = fmin;
}

//======================================  Set the maximum integration frequency
void
ranger::setFmax(double fmax) {
   mFmax = fmax;
}

//======================================  calculate and set the gw template
void
ranger::setGWType(const std::string& name) {
   mGWType = name;
   PSD gw_template = get_template(name);
   setWeightZ0(gw_template);
}

//======================================  Set the histogram bins.
void
ranger::setHistogramBins(int nBin, double low, double high) {
   mRangeHisto.SetBinLowEdges(nBin, low, high);
}

//======================================  Set the integrand monitor channel name
void
ranger::setIntegrandMonitor(const std::string& name) {
   mIntegrandMonitor = name;
}

//======================================  Set the monitor history length
void
ranger::setHistoryLength(Interval len) {
   mHistoryLength = len;
   mRangeHistory.setMaxLen(len);
   mRedShiftHistory.setMaxLen(len);
}

//======================================  Set the Psd monitor channel name
void
ranger::setPsdMonitor(const std::string& name) {
   mPsdMonitor = name;
}

//======================================  Set the psd estimate tool
void
ranger::setPSDTool(const psd_estimate& est) {
   mPsdTool.reset(est.clone());
   mWindowLength = est.stride();
};

//======================================  Set the range trend channel name
void
ranger::setRangeHistoName(const std::string& name) {
   mRangeHistoName = name;
};

//======================================  Set the range monitor channel name
void
ranger::setRangeMonitor(const std::string& name) {
   mRangeMonitor = name;
};

//======================================  Set the range trend channel name
void
ranger::setRangeTrend(const std::string& name) {
   mRangeTrend = name;
};

//======================================  Set the red-shift trend channel name
void
ranger::setRedShiftTrend(const std::string& name) {
   mRedShiftTrend = name;
};

//======================================  Set the red-shift monitor channel name
void
ranger::setRedShiftMonitor(const std::string& name) {
   mRedShiftMonitor = name;
};

//======================================  The sigma threshold
void
ranger::setThreshold(double sigma) {
   mSigThreshold = sigma;
};

//======================================  Set the signal psd
void
ranger::setWeightZ0(const containers::PSD& psd) {
   mWeightZ0 = psd;
}
