/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "ap-channels.hh"
#include "TSeries.hh"
#include "DVecType.hh"
#include "lcl_array.hh"
#include <sstream>
#include <algorithm>

#include "poissonge.icc"

using namespace std;

//======================================  Channel constructor
channels::channels(const std::string& chan, size_t N)
   : _alloc(N), _name(chan), _reject(0), _debug(0)
{
}

//=======================================  Destructor
channels::~channels(void) {
}

//=======================================  print bin id and limits
std::string
channels::bin_id(size_t id) const {
   ostringstream binid;
   binid << id << " {" << _limits[id] << ", " << _limits[id+1] << "}";
   return binid.str();
} 

//=======================================  Append data
void 
channels::append(const TSeries& ts) {
   if (ts.empty()) return;
   if (_value.empty()) {
      DVector* dv = ts.refDVect()->clone();
      dv->reserve(_alloc);
      _value.setData(ts.getStartTime(), ts.getTStep(), dv);
   } else {
      _value.Append(ts);
   }
}

//=======================================  flush overlap
void
channels::next(Interval dt) {
   _value.eraseStart(dt);
}

void
channels::reset(void) {
   _value.Clear();
}

//===================================== set the debug print level.
void
channels::set_debug(int deb) {
   _debug = deb;
}

//=======================================  Set the limits
void 
channels::set_limits(size_t M) {
   DVectF cdata(*_value.refDVect());
   DVectF data = cdata;
   size_t N=data.size();
   std::sort(data.refTData(), data.refTData()+N);
   _limits.resize(M+1, 0);

   //-----------------------------------  Pick up the rank limit values.
   _limits[0] = data[0];
   _limits[M] = data[N-1];
   size_t div = N / M;
   size_t ndup = 0;
   for (size_t i=1; i<M; i++) {
      size_t inx = i * div;
      _limits[i] = data[inx];
      if (_limits[i] == data[inx-1]) ndup++;
   }
   if (ndup) {
      if (_debug > 1) {
	 cout << "warning - " << ndup << " slices of channel: " << name()
	      << " may share values." << endl;
      }
      set_reject(2);
   }

   //-------------------------------------  Set up the rank Series.
   DVectI rvect(N);
   size_t rinx = 0;
   for (size_t i=0; i<N; i++) {
      float cdi = cdata[i];
      while (cdi < _limits[rinx] || cdi >= _limits[rinx+1]) {
	 if (cdi < _limits[rinx]) {
	    if (!rinx) break;
	    rinx--;
	 } else if (cdi >= _limits[rinx+1]) {
	    if (rinx == M-1) break;
	    rinx++;
	 }
      }
      rvect[i] = rinx;
   }
   _rank.setData(_value.getStartTime(), _value.getTStep(), rvect.clone());
}

//===================================== set the reject bit
void
channels::set_reject(int rej) {
   _reject = rej;
}

//=====================================  Test the trigger rates per rank
bool
channels::test1D(const TSeries& ts, double logCL, chan_stat& stat,
		 bool reject_outliers) {
   size_t N = _rank.getNSample();
   double nTrig  = ts.refDVect()->VSum(0, N);
   double perbin = nTrig/double(N);
   size_t M = 16;
   bool   fail = false;

   //----------------------------------  Set up count arrays
   lcl_array<size_t> kval(M);
   lcl_array<size_t> tval(M);
   memset(kval.get(), 0, M*sizeof(size_t));
   memset(tval.get(), 0, M*sizeof(size_t));

   //----------------------------------  Count entries and triggers
   const DVectI& dvr(dynamic_cast<const DVectI&>(*_rank.refDVect()));
   const DVectI& dvt(dynamic_cast<const DVectI&>(*ts.refDVect()));
   for (size_t i=0; i<N; i++) {
      int k = dvr[i];
      kval[k]++;
      tval[k] += dvt[i];
   }

   //----------------------------------  Check the stats
   int wi=-1;
   double wx = logCL;
   double wav; 
   for (size_t i=0; i<M; i++) {
      double average = perbin * kval[i];
      double pois    = poissonge(average, tval[i]);
      if (pois <= 0) continue;
      pois = log(pois);
      if (pois < wx) {
         wi = i;
         wx = pois;
         wav = average;
      }
   }

   //----------------------------------  Fill the statistics structure
   stat.min_index = wi;
   stat.min_likelihood = wx;
   stat.trigs_found = tval[wi];
   stat.trigs_expected = wav;

   //----------------------------------  Tell the world if this is significant
   if (wi >= 0) {
      if (_debug > 1) {
	 cout << name() << "[" << wi << "] N=" << tval[wi] 
	      << " expect=" << wav << " pois=" << wx << " (log " << exp(wx)
	      << ")" << endl;
	 cout << "            range " << wi << ": " << _limits[wi] << " - "
	      << _limits[wi+1] << endl;
      }
      if (reject_outliers) {
	 set_reject(3);
	 fail = true;
      }
   }
   return fail;
}
